// Crimson Fields -- a game of tactical warfare
// Copyright (C) 2000-2003 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

////////////////////////////////////////////////////////////////////////
// history.cpp
//
// History:
//  05-06-2001 - created
//  01-05-2002 - removed destructor
////////////////////////////////////////////////////////////////////////

#include "SDL_endian.h"

#include "history.h"
#include "options.h"

extern Options CFOptions;

////////////////////////////////////////////////////////////////////////
// NAME       : HistEvent::HistEvent
// DESCRIPTION: Create a new event for the turn history.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

HistEvent::HistEvent( void ) {
  type = HIST_DONE;
}

////////////////////////////////////////////////////////////////////////
// NAME       : HistEvent::HistEvent
// DESCRIPTION: Load an event from a file.
// PARAMETERS : file - SDL_RWops file descriptor
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

HistEvent::HistEvent( SDL_RWops *file ) {
  SDL_RWread( file, &type, 1, 1 );
  for ( int i = 0; i < 4; ++i ) data[i] = SDL_ReadLE16( file );
}

////////////////////////////////////////////////////////////////////////
// NAME       : HistEvent::Save
// DESCRIPTION: Save an event to a file.
// PARAMETERS : file - SDL_RWops file descriptor
// RETURNS    : 0 on success, non-0 otherwise
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int HistEvent::Save( SDL_RWops *file ) const {
  SDL_RWwrite( file, &type, 1, 1 );
  for ( int i = 0; i < 4; ++i ) SDL_WriteLE16( file, data[i] );
  return 0;
}


////////////////////////////////////////////////////////////////////////
// NAME       : History::History
// DESCRIPTION: Load a History from a file.
// PARAMETERS : file - open SDL_RWops descriptor
//              num  - number of recorded events in file
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

History::History( SDL_RWops *file, unsigned short num ) {
  unsigned short num_units, i;

  num_units = SDL_ReadLE16( file );

  for ( i = 0; i < num; ++i ) {
    events.AddTail( new HistEvent( file ) );
  }

  for ( i = 0; i < num_units; ++i ) {
    Unit *u = Gam->LoadUnit( file, true );
    if ( u ) units.AddTail( u );
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::History
// DESCRIPTION: Create a new (empty) History from the current game.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

History::History( void ) {
  // create the unit copies
  Unit *dummy, *u = static_cast<Unit *>( Gam->UnitsList()->Head() );
  while ( u ) {
    dummy = new Unit( *u );
    dummy->SetFlags( U_DUMMY );
    units.AddTail( dummy );
    u = static_cast<Unit *>( u->Next() );
  }
}


////////////////////////////////////////////////////////////////////////
// NAME       : History::Save
// DESCRIPTION: Save a history to file.
// PARAMETERS : file - open SDL_RWops file descriptor
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int History::Save( SDL_RWops *file ) const {
  unsigned short num = events.CountNodes();
  SDL_WriteLE16( file, num );
  if ( num == 0 ) return 0;

  num = units.CountNodes();
  SDL_WriteLE16( file, num );

  // save events
  HistEvent *he = static_cast<HistEvent *>( events.Head() );
  while ( he ) {
    he->Save( file );
    he = static_cast<HistEvent *>( he->Next() );
  }

  // save units
  Unit *u = static_cast<Unit *>( units.Head() );
  while ( u ) {
    u->Save( file );
    u = static_cast<Unit *>( u->Next() );
  }
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::Replay
// DESCRIPTION: Play all events that have been stored during the last
//              turn.
// PARAMETERS : mapwin - pointer to MapWindow
// RETURNS    : -
//
// HISTORY
//   12-08-2002 - in a HIST_MOVE event of a transport also check whether
//                we moved out of a transport (not only building)
//   23-08-2002 - adapt HIST_MOVE to modified Game::MoveUnit() methods
//              - fix mem leak in HIST_COMBAT
////////////////////////////////////////////////////////////////////////

void History::Replay( MapWindow *mapwin ) {
  if ( events.IsEmpty() ) return;

  View *view = mapwin->GetView();
  MapView *mv = mapwin->GetMapView();
  Map *map = mv->GetMap();
  ProgressWindow *progwin = NULL;
  MessageWindow *msgw;
  HistEvent *he = static_cast<HistEvent *>( events.Head() );

  bool abort = false;
  bool replay = CFOptions.GetTurnReplay();

  if ( replay ) {
    List units_bak;
    BeginReplay( units_bak, map );

    unsigned short count = events.CountNodes();

    // undo the map tile changes which have been recorded this turn
    while ( he && (he->type == HIST_TILE_INTERNAL) ) {
      map->SetHexType( he->data[1], he->data[2], he->data[0] );
      he->type = HIST_DONE;
      he = static_cast<HistEvent *>( he->Next() );
      --count;
    }

    mv->Enable();
    mapwin->Draw();
    mapwin->Show();
    progwin = new ProgressWindow( 5,
              mapwin->Height() - view->SmallFont()->Height() - 30,
              mapwin->Width() / 2, view->SmallFont()->Height() + 20,
              0, count, WIN_PROG_ABORT, view );

    Unit *lastunit = NULL;
    Point lastpos( -1, -1 );
    Rect upd;

    while ( he && !abort ) {
      Unit *u;

      switch ( he->type ) {
        case HIST_MOVE:              // move unit
          u = Gam->GetUnitByID( he->data[0] );
          if ( u ) {
            if ( u != lastunit ) {
              if ( lastunit && lastunit->IsTransport() && !map->GetMapObject( lastpos ) ) {
                // we must adjust the position of carried units manually because
                // those pseudo-transports can't handle this themselves
                for ( Unit *carry = static_cast<Unit *>( Gam->UnitsList()->Head() );
                      carry; carry = static_cast<Unit *>( carry->Next() ) ) {
                  if ( carry->Position() == lastpos )
                    carry->SetPosition( lastunit->Position().x, lastunit->Position().y );
                }
              }
              lastunit = u;
              lastpos = u->Position();

              SDL_Delay( DEFAULT_DELAY );
            }

            if ( !mv->HexVisible( u->Position() ) ) {
              mv->CenterOnHex( u->Position() );
              mapwin->Show( *mv );
            }

            if ( u->IsSheltered() ) {
              if ( map->GetUnit( u->Position() ) ) u->UnsetFlags( U_SHELTERED );
              else map->GetBuilding( u->Position() )->RemoveUnit( u );
            } else {
              map->SetUnit( NULL, u->Position() );
              mv->UpdateHex( u->Position() );
            }

            Gam->MoveUnit( u, (Direction)he->data[1] );
            map->SetUnit( u, u->Position() );
            mapwin->Show( mv->UpdateHex( u->Position() ) );
          }
          break;
        case HIST_ATTACK:
          u = Gam->GetUnitByID( he->data[0] );
          if ( u ) {
            if ( u != lastunit ) SDL_Delay( DEFAULT_DELAY );
            if ( !mv->HexVisible( u->Position() ) ||
                 !mv->HexVisible( Point(he->data[1], he->data[2]) ) ) {
              mv->CenterOnHex( Point((u->Position().x + he->data[1]) / 2,
                                (u->Position().y + he->data[2]) / 2) );
              mapwin->Show( *mv );
            }
            mv->SetCursorImage( IMG_CURSOR_HIGHLIGHT );
            upd = mv->SetCursor( u->Position() );
            mapwin->Show( upd );
            mv->SetCursorImage( IMG_CURSOR_ATTACK );
            upd = mv->SetCursor( Point(he->data[1], he->data[2]) );
            mapwin->Show( upd );
            mv->SetCursor( Point(-1,-1) );
            SDL_Delay( DEFAULT_DELAY );
            upd = mv->UpdateHex( u->Position() );
            mapwin->Show( upd );
            upd = mv->UpdateHex( Point(he->data[1], he->data[2]) );
            mapwin->Show( upd );
            lastunit = u;
          }
          break;
        case HIST_COMBAT: {
          Combat cmb( Gam->GetUnitByID( he->data[0] ),
                      Gam->GetUnitByID( he->data[1] ) );
          cmb.aamod = (signed char)he->data[3];
          cmb.ddmod = (signed char)he->data[2];
          CombatWindow *cwin = cmb.Resolve( mapwin, view );
          if ( cwin ) {
            cwin->EventLoop();
            view->CloseWindow( cwin );
          }
          break; }
        case HIST_TILE:
          map->SetHexType( he->data[1], he->data[2], he->data[0] );
          upd = mv->UpdateHex( Point(he->data[1], he->data[2]) );
          mapwin->Show( upd );
          SDL_Delay( DEFAULT_DELAY );
          break;
        case HIST_MSG:
          msgw = new MessageWindow(
            (he->data[0] != -1) ? Gam->GetMessage( he->data[0] ) : "Message",
            Gam->GetMessage( he->data[1] ), view );
          msgw->EventLoop();
          view->CloseWindow( msgw );
          break;
        default:
          break;
      }
      he->type = HIST_DONE;
      he = static_cast<HistEvent *>( he->Next() );
      progwin->Advance( 1 );
      abort = progwin->Aborted();
    }

    EndReplay( units_bak, map );
  }

  if ( replay && !abort ) SDL_Delay( DEFAULT_DELAY * 2 );
  else {
    // if replay has been stopped or disabled we still need to revert
    // all changes to the map and display the cached messages
    while ( he ) {
      if ( he->type == HIST_TILE ) map->SetHexType( he->data[1], he->data[2], he->data[0] );
      else if ( he->type == HIST_MSG ) {
        msgw = new MessageWindow(
          (he->data[0] != -1) ? Gam->GetMessage( he->data[0] ) : "Message",
          Gam->GetMessage( he->data[1] ), view );
        msgw->EventLoop();
        view->CloseWindow( msgw );
      }
      he = static_cast<HistEvent *>( he->Next() );
    }
  }

  if ( progwin ) view->CloseWindow( progwin );
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::BeginReplay
// DESCRIPTION: Prepare for viewing the last turn replay. We must store
//              the current list of units in a safe place and replace
//              them by their "historical" versions.
// PARAMETERS : backup - list to store current unit list in
//              map    - map
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void History::BeginReplay( List &backup, Map *map ) {
  List *gam_units = Gam->UnitsList();
  Unit *u;

  while ( !gam_units->IsEmpty() ) {
    u = static_cast<Unit *>( gam_units->RemHead() );
    backup.AddTail( u );
    if ( !u->IsSheltered() ) map->SetUnit( NULL, u->Position() );
  }

  while ( !units.IsEmpty() ) {
    u = static_cast<Unit *>( units.RemHead() );
    gam_units->AddTail( u );
    if ( !u->IsSheltered() ) {
      map->SetUnit( u, u->Position() );
      if ( u->IsMine() ) u->SetMoves( MCOST_MIN );
    }
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::EndReplay
// DESCRIPTION: End replay mode and initiate the next turn. Remove the
//              historical units and put the real ones back in place.
// PARAMETERS : backup - list containing the real units
//              map    - map
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void History::EndReplay( List &backup, Map *map ) {
  List *gam_units = Gam->UnitsList();
  Unit *u;

  while ( !gam_units->IsEmpty() ) {
    u = static_cast<Unit *>( gam_units->RemHead() );
    if ( !u->IsSheltered() ) map->SetUnit( NULL, u->Position() );
    delete u;
  }

  while ( !backup.IsEmpty() ) {
    u = static_cast<Unit *>( backup.RemHead() );
    gam_units->AddTail( u );
    if ( !u->IsSheltered() ) map->SetUnit( u, u->Position() );
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::RecordMoveEvent
// DESCRIPTION: Record the movement of a unit from one hex to another,
//              neighboring hex.
// PARAMETERS : u   - unit to be moved
//              dir - direction to move the unit in
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int History::RecordMoveEvent( const Unit &u, Direction dir ) {
  HistEvent *he = new HistEvent;

  he->type = HIST_MOVE;
  he->data[0] = u.ID();
  he->data[1] = dir;
  events.AddTail( he );
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::RecordAttackEvent
// DESCRIPTION: Record an attack being inititated.
// PARAMETERS : u      - unit to inititate the attack
//              target - unit being attacked
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int History::RecordAttackEvent( const Unit &u, const Unit &target ) {
  HistEvent *he = new HistEvent;

  he->type = HIST_ATTACK;
  he->data[0] = u.ID();
  he->data[1] = target.Position().x;
  he->data[2] = target.Position().y;

  events.AddTail( he );
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::RecordCombatEvent
// DESCRIPTION: Record the results of a combat.
// PARAMETERS : combat - combat data
//              loss1  - attackers' casualties
//              loss2  - defenders' casualties
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int History::RecordCombatEvent( const Combat &combat, unsigned char loss1,
                                unsigned char loss2 ) {
  HistEvent *he = new HistEvent;

  he->type = HIST_COMBAT;
  he->data[0] = combat.c_att->ID();
  he->data[1] = combat.c_def->ID();
  he->data[2] = loss1;
  he->data[3] = loss2;

  events.AddTail( he );
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::RecordTileEvent
// DESCRIPTION: Record a modification of the map. This event is special
//              in so far as two separate events are created for each
//              event. One allows for the initial map to be created from
//              the map at the end of the turn, the other represents the
//              actual map change.
// PARAMETERS : tile - tile type being set
//              old  - tile type being replaced
//              dx   - x hex being modified
//              dy   - y hex being modified
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int History::RecordTileEvent( unsigned short tile, unsigned short old,
                              short dx, short dy ) {
  HistEvent *he = new HistEvent;

  // first record the actual change
  he->type = HIST_TILE;
  he->data[0] = tile;
  he->data[1] = dx;
  he->data[2] = dy;
  events.AddTail( he );

  // now create the map state initializer event and put it at the
  // head of the list
  he = new HistEvent;
  he->type = HIST_TILE_INTERNAL;
  he->data[0] = old;
  he->data[1] = dx;
  he->data[2] = dy;
  events.AddHead( he );
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : History::RecordMsgEvent
// DESCRIPTION: Record a message for the next player.
// PARAMETERS : title - title to use for the message window. If -1,
//                      used title will be "Message".
//              msg   - index of the message
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int History::RecordMsgEvent( short title, unsigned short msg ) {
  HistEvent *he = new HistEvent;

  he->type = HIST_MSG;
  he->data[0] = title;
  he->data[1] = msg;
  he->data[2] = he->data[3] = -1;

  events.AddTail( he );
  return 0;
}

