// Crimson Fields -- a hex-based game of tactical warfare
// Copyright (C) 2000-2003 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

/////////////////////////////////////////////////////////////////////
// main.cpp -- Crimson Fields
//
// History:
//  23-03-2000 - created
//  27-04-2000 - support level files
//  01-03-2001 - initial public release
//  07-03-2001 - include string.h to fix compilation errors
//  25-03-2001 - added main menu to select level
//  02-04-2001 - added glob var AllowUpd to disallow display surface
//               updates if necessary
//  22-04-2001 - display depth now 16 bpp to use true "alpha fog"
//             - added sound support
//  04-05-2001 - reintroduced --level command line parameter
//  16-05-2001 - added --version and --help command line parameters
//  26-05-2001 - removed AllowUpd, moved into View class
//  19-06-2001 - added event filter callback function
//  25-08-2002 - don't init audio subsystem if sound is off
//  03-09-2002 - make signed chars explicit where expected
//  08-08-2003 - added support for WindowsCE
/////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <string.h>

#ifndef _WIN32_WCE
# include <time.h>
# include <fstream.h>
#endif

#include "SDL.h"
#include "SDL_endian.h"

#include "view.h"
#include "game.h"
#include "misc.h"
#include "fileio.h"
#include "initwindow.h"
#include "globals.h"
#include "sound.h"
#include "options.h"

// global vars
Game *Gam;
Image *Images[NUM_IMAGES] = { NULL };

// global options, can be accessed from everywhere
Options CFOptions;

struct GUIOptions {
  short px_width;
  short px_height;
  short bpp;
  bool sfx;
  bool music;
  unsigned char sfx_vol;
  unsigned char music_vol;
  unsigned long sdl_flags;
  const char *level;
};

// local vars
static View *display = NULL;

// local function prototypes
static void parse_options( int argc, char **argv, GUIOptions &opts );
static void print_usage( char *prog );
static int init_data( View *display, GUIOptions &opts );
static void load_settings( GUIOptions &opts );
static void save_settings( View *display );
static GUI_Status event_filter( SDL_Event &event, Window *window );
static void do_exit( void );

#ifdef _WIN32_WCE
static bool init_wince( struct GUIOptions &opts );
# define time(n) GetTickCount()
#endif

// Start of program functions //////////////////////////////////////////

#ifdef _WIN32_WCE
int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPWSTR szCmdLine, int sw) {
#else
int main( int argc, char **argv ) {
#endif
  struct GUIOptions guiopts = { 800, 600, DISPLAY_BPP, true, true,
                    MIX_MAX_VOLUME*3/4, MIX_MAX_VOLUME/2, SDL_HWSURFACE, NULL };

#ifdef _WIN32_WCE
  if ( !init_wince( guiopts ) ) return 0;
#else
  parse_options( argc, argv, guiopts );
#endif

  if ( SDL_Init( SDL_INIT_VIDEO ) < 0 ) {
    fprintf( stderr, "Couldn't initialize: %s\n", SDL_GetError() );
    return 1;
  }

  // set main window title
  SDL_WM_SetCaption( PROGRAMNAME, CF_SHORTNAME );
  SDL_EnableKeyRepeat( SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL );

  display = new View( guiopts.px_width, guiopts.px_height, guiopts.bpp,
                      guiopts.sdl_flags );
  if ( !display->s_surface ) fprintf( stderr, "Couldn't set video mode: %s\n", SDL_GetError() );
  else {

    if ( !init_data( display, guiopts ) ) {
      srand( time(0) );			// initialize random number generator

      display->SetEventFilter( event_filter );

      // create options and save games directories if possible
      create_config_dir();

      // only open intro screen if the user didn't supply a level on the command line
      int intro = 1;
      if ( guiopts.level ) {
        Gam = new Game( display );
        intro = Gam->Load( guiopts.level );
        if ( !intro ) {
          if ( !Gam->GetLevelPassword() ||
               Gam->CheckPassword( "Access restricted", "Enter access code",
                                   Gam->GetLevelPassword(), false ) ) Gam->StartTurn();
          else intro = 1;
        }
      }

      GUI_Status status;
      do {
        if ( intro ) {
          display->Refresh();
          TitleWindow *twin = new TitleWindow( display );
          new InitWindow( display, twin );
        } else intro = 1;

        do {
          status = display->HandleEvents();
        } while ( (status != GUI_QUIT) && (status != GUI_RESTART) );

        delete Gam;
        Gam = NULL;
        display->CloseAllWindows();
      } while ( status != GUI_QUIT );
    }
  }

  do_exit();
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : parse_options
// DESCRIPTION: Process any options given to the program on the command
//              line.
// PARAMETERS : argc - argument count
//              argv - pointer to array of arguments
//              opts - buffer to store GUI and audio options; should be
//                     initialized with defaults before calling this
//                     method
// RETURNS    : -
//
// HISTORY
//   18-06-2001 - fullscreen option now takes an argument
//   23-09-2001 - added --sound command line parameter
////////////////////////////////////////////////////////////////////////

// parse command line arguments
void parse_options( int argc, char **argv, GUIOptions &opts ) {
  load_settings( opts );

  while ( argc > 1 ) {
    --argc;

    if (strcmp(argv[argc-1], "--width") == 0) {
      opts.px_width = atoi(argv[argc]);
    } else if (strcmp(argv[argc-1], "--height") == 0) {
      opts.px_height = atoi(argv[argc]);
    } else if (strcmp(argv[argc-1], "--level") == 0) {
      opts.level = argv[argc];
    } else if (strcmp(argv[argc-1], "--fullscreen") == 0) {
      if ( atoi( argv[argc] ) ) opts.sdl_flags |= SDL_FULLSCREEN;
      else opts.sdl_flags &= ~SDL_FULLSCREEN;
    } else if (strcmp(argv[argc-1], "--sound") == 0) {
      if ( atoi( argv[argc] ) ) opts.sfx = opts.music = true;
      else opts.sfx = opts.music = false;
    } else {
      if (strcmp(argv[argc], "--version") == 0)
        fprintf( stdout, PROGRAMNAME" "VERSION"\n" );
      else print_usage( argv[0] );
      exit ( 0 );
    }
    --argc;
  }
  if ( opts.px_width < MIN_XRES ) opts.px_width = MIN_XRES;
  if ( opts.px_height < MIN_YRES ) opts.px_height = MIN_YRES;
}

////////////////////////////////////////////////////////////////////////
// NAME       : print_usage
// DESCRIPTION: Print a usage message to stdout.
// PARAMETERS : prog - program name as given on the command line
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void print_usage( char *prog ) {
  fprintf( stdout, "Usage: %s [options]\n\n"
                   "Available options:\n"
                   "  --level <level>      load level or save file\n"
                   "  --width <width>      set screen width\n"
                   "  --height <height>    set screen height\n"
                   "  --fullscreen <1|0>   enable/disable fullscreen mode\n"
#ifndef DISABLE_SOUND
                   "  --sound <1|0>        enable/disable sound\n"
#endif
                   "  --help               display this help and exit\n"
                   "  --version            output version information and exit\n",
        prog );
}

////////////////////////////////////////////////////////////////////////
// NAME       : init_data
// DESCRIPTION: Initialize the system icons surface and the game fonts.
// PARAMETERS : view  - display surface
//              opts  - contains the audio settings required to properly
//                      initialize the sound subsystem
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int init_data( View *view, GUIOptions &opts ) {
  string datpath( get_data_dir() );
  datpath.append( CF_DATFILE );
  SDL_RWops *datfile = SDL_RWFromFile( datpath.c_str(), "rb" );
  if ( !datfile ) {
    fprintf( stderr, "Error: Couldn't open '%s'\n", datpath.c_str() );
    return -1;
  }

  Surface *icons = new Surface;			// load icons surface
  if ( icons->LoadImageData( datfile ) ) {
    fprintf( stderr, "Error reading data file\n" );
    SDL_RWclose( datfile );
    delete icons;
    return -1;
  }

  Font *f1 = new Font( datfile );	// load fonts
  Font *f2 = new Font( datfile );

  SDL_RWclose( datfile );

  Audio::InitSfx( opts.sfx, opts.sfx_vol );		// load sound effects
  Audio::InitMusic( opts.music, opts.music_vol );

  Images[ICON_SPEED] = new Image( icons, 0 * ICON_WIDTH, 0 * ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT );
  Images[ICON_DEFENCE] = new Image( icons, 1 * ICON_WIDTH, 0 * ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT );
  Images[ICON_GROUND] = new Image( icons, 2 * ICON_WIDTH, 0 * ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT );
  Images[ICON_AIRCRAFT] = new Image( icons, 3 * ICON_WIDTH, 0 * ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT );
  Images[ICON_SHIP] = new Image( icons, 4 * ICON_WIDTH, 0 * ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT );
  Images[ICON_CRYSTALS] = new Image( icons, 5 * ICON_WIDTH, 0 * ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT );
  Images[ICON_HEX_BG] = new Image( icons, 1 * ICON_WIDTH, 1 * ICON_HEIGHT, ICON_WIDTH, ICON_HEIGHT );
  for ( int i = 0; i <= XP_MAX_LEVEL; ++i )
    Images[ICON_XP_BASE+i] = new Image( icons, 64 + i * XP_ICON_WIDTH, 32, XP_ICON_WIDTH, XP_ICON_HEIGHT );
  Images[ICON_BLD_WORKSHOP] = new Image( icons, 64 + 7 * XP_ICON_WIDTH, 32, XP_ICON_WIDTH, XP_ICON_HEIGHT );
  Images[ICON_BLD_FACTORY] = new Image( icons, 64 + 8 * XP_ICON_WIDTH, 32, XP_ICON_WIDTH, XP_ICON_HEIGHT );

  view->SetSmallFont( f1 );
  view->SetLargeFont( f2 );
  icons->DisplayFormat();
  view->SetSystemIcons( icons );
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : load_settings
// DESCRIPTION: Read default display settings from the crimsonrc file.
// PARAMETERS : opts - buffer to store the settings. These should
//                     already be initialized with some defaults in
//                     case the rc file doesn't exist or this function
//                     fails.
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void load_settings( GUIOptions &opts ) {
  string crimsonrc( get_config_dir() );
  crimsonrc.append( CRIMSONRC );

  FILE *fp = fopen( crimsonrc.c_str(), "r" );
  if ( fp ) {
    char linebuf[256], *val;
    unsigned short linecnt = 0;

    while ( fgets( linebuf, 255, fp ) ) {
      ++linecnt;
      if ( (linebuf[0] != '#') && (linebuf[0] != '\n') ) {
        val = strchr( linebuf, ' ' );
        if ( val ) {
          while ( *val == ' ' ) ++val;

          if ( !strncmp( linebuf, "width", 5 ) ) opts.px_width = atoi(val);
          else if ( !strncmp( linebuf, "height", 6 ) ) opts.px_height = atoi(val);
          else if ( !strncmp( linebuf, "fullscreen", 10 ) ) {
            if ( atoi(val) != 0 ) opts.sdl_flags |= SDL_FULLSCREEN;
          } else if ( !strncmp( linebuf, "sfxvol", 6 ) ) opts.sfx_vol = atoi(val);
          else if ( !strncmp( linebuf, "musicvol", 8 ) ) opts.music_vol = atoi(val);
          else if ( !strncmp( linebuf, "sfx", 3 ) ) opts.sfx = (atoi(val) != 0);
          else if ( !strncmp( linebuf, "music", 5 ) ) opts.music = (atoi(val) != 0);
          else if ( !strncmp( linebuf, "showdamage", 10 ) ) CFOptions.SetDamageIndicator( atoi(val) != 0 );
          else if ( !strncmp( linebuf, "showreplay", 10 ) ) CFOptions.SetTurnReplay( atoi(val) != 0 );
          else fprintf( stderr, "warning: unrecognized config option in line %d\n", linecnt );
        }
      }
    }
    fclose( fp );
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : save_settings
// DESCRIPTION: Save current display settings to the crimsonrc file.
// PARAMETERS : display - pointer to display
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void save_settings( View *display ) {
  string crimsonrc( get_config_dir() );
  crimsonrc.append( CRIMSONRC );

  FILE *fp = fopen( crimsonrc.c_str(), "w" );
  if ( fp ) {
    fprintf( fp, "width %d\n", display->Width() );
    fprintf( fp, "height %d\n", display->Height() );
    fprintf( fp, "fullscreen %d\n", display->IsFullScreen() );
    fprintf( fp, "sfx %d\n", Audio::GetSfxState() );
    fprintf( fp, "music %d\n", Audio::GetMusicState() );
    fprintf( fp, "sfxvol %d\n", Audio::GetSfxVolume() );
    fprintf( fp, "musicvol %d\n", Audio::GetMusicVolume() );
    fprintf( fp, "showdamage %d\n", CFOptions.GetDamageIndicator() );
    fprintf( fp, "showreplay %d\n", CFOptions.GetTurnReplay() );
    fclose( fp );
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : event_filter
// DESCRIPTION: This is the global event filter function. It is hooked
//              to the display and called everytime the event handler
//              receives an event.
// PARAMETERS : event  - event received by the event handler
//              window - pointer to the currently active window
// RETURNS    : GUI_Status; if the filter returns GUI_NONE the event
//              handler will not pass the event to its windows, but
//              silently drop it.
//
// HISTORY
//   13-08-2001 - F11 toggles sound on/off
//   29-04-2003 - moved fullscreen toggle to View
////////////////////////////////////////////////////////////////////////

GUI_Status event_filter( SDL_Event &event, Window *window ) {
  GUI_Status rc = GUI_OK;

  if ( event.type == SDL_KEYDOWN ) {
    rc = GUI_NONE;

    switch ( event.key.keysym.sym ) {
    case SDLK_F11:            // toggle sound
      Audio::ToggleSfxState();
      break;
    default:
      rc = GUI_OK;            // send to windows
    }
  } else if ( event.type == SDL_QUIT ) do_exit();

  return rc;
}

////////////////////////////////////////////////////////////////////////
// NAME       : init_wince
// DESCRIPTION: Initialize on Windows CE.
// PARAMETERS : opts - program options buffer (containing defaults)
// RETURNS    : TRUE if initialization succeeded, FALSE otherwise
//
// HISTORY
////////////////////////////////////////////////////////////////////////

#ifdef _WIN32_WCE
bool init_wince( struct GUIOptions &opts ) {
  HWND hwnd = FindWindow( _T("SDL_app"), _T(PROGRAMNAME) );
  if ( hwnd ) {
    SetForegroundWindow( hwnd );
    ShowWindow( hwnd, SW_SHOW );
    return false;
  }

  load_settings( opts );

  unsigned short maxw = GetSystemMetrics(SM_CXSCREEN); 
  unsigned short maxh = GetSystemMetrics(SM_CYSCREEN); 

  if( opts.px_width > maxw ) opts.px_width = maxw;
  if( opts.px_height > maxh ) opts.px_height = maxh;

  opts.sdl_flags |= SDL_FULLSCREEN;
  return true;
}
#endif

////////////////////////////////////////////////////////////////////////
// NAME       : do_exit
// DESCRIPTION: Free all resources and exit the program.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void do_exit( void ) {
  Audio::ShutdownSfx();
  Audio::ShutdownMusic();

  if ( display ) {
    save_settings( display );
    delete display;
  }

  for ( int i = 0; i < NUM_IMAGES; ++i ) delete Images[i];

  SDL_Quit();
  exit( 0 );
}

