// Crimson Fields -- a game of tactical warfare
// Copyright (C) 2000-2003 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

/////////////////////////////////////////////////////////////////////////
// unit.cpp
//
// History:
//  04-12-2000 - created
//  12-08-2001 - added sounds to unit definition
//  16-03-2002 - adapt to modified UnitType class
////////////////////////////////////////////////////////////////////////

#include "SDL_endian.h"

#include "unit.h"
#include "hexsup.h"

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Unit
// DESCRIPTION: Create a new unit.
// PARAMETERS : type   - pointer to a unit type definition
//              player - pointer to the player this unit belongs to
//              id     - unique unit identifier
//              x      - horizontal position on map
//              y      - vertical position on map
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Unit::Unit( const UnitType *type, Player *player, unsigned short id, short x, short y ) :
      MapObject( MO_UNIT, type->Name() ) {
  u_type = type;
  u_id = id;
  u_pos.x = x;
  u_pos.y = y;

  u_flags = type->Flags();
  u_moves = 0;
  u_facing = NORTH;
  u_group = MAX_GROUP_SIZE;
  u_xp = 0;

  u_target.x = u_target.y = 0;

  u_pid = PLAYER_NONE;
  u_player = NULL;

  SetOwner( player );
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Unit
// DESCRIPTION: Load a unit from a data file.
// PARAMETERS : file   - descriptor of an open data file
//              type   - pointer to a unit type definition
//              player - pointer to the player this unit belongs to
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Unit::Unit( SDL_RWops *file, const UnitType *type, Player *player ) :
      MapObject( MO_UNIT, type->Name() ) {
  u_pos.x = SDL_ReadLE16( file );
  u_pos.y = SDL_ReadLE16( file );
  u_flags = SDL_ReadLE32( file );
  u_id = SDL_ReadLE16( file );

  SDL_RWread( file, &u_moves, 1, 4 );

  u_target.x = SDL_ReadLE16( file );
  u_target.y = SDL_ReadLE16( file );

  u_type = type;
  u_pid = PLAYER_NONE;
  u_player = NULL;

  SetOwner( player );
}


////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Save
// DESCRIPTION: Save the unit to a data file.
// PARAMETERS : file - descriptor of the save file
// RETURNS    : 0 on succes, non-zero on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int Unit::Save( SDL_RWops *file ) const {
  unsigned char typeno = u_type->ID();
  SDL_RWwrite( file, &typeno, 1, 1 );
  SDL_RWwrite( file, &u_pid, 1, 1 );

  SDL_WriteLE16( file, u_pos.x );
  SDL_WriteLE16( file, u_pos.y );
  SDL_WriteLE32( file, u_flags );
  SDL_WriteLE16( file, u_id );

  SDL_RWwrite( file, &u_moves, 1, 4 );

  SDL_WriteLE16( file, u_target.x );
  SDL_WriteLE16( file, u_target.y );

  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::AwardXP
// DESCRIPTION: Raise the unit's experience level. This will improve its
//              chances to survive a battle and inflict more damage to
//              the enemy.
// PARAMETERS : xp - number of experience points to give; a unit
//                   advances to a new experience level every 3 points
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void Unit::AwardXP( unsigned char xp ) {
  // XP_MAX_LEVEL is the maximum experience level a unit can reach,
  // i.e. XP_MAX_LEVEL * XP_PER_LEVEL is the upper limit in points
  u_xp = MIN( u_xp + xp, XP_MAX_LEVEL * XP_PER_LEVEL );
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::SetOwner
// DESCRIPTION: Change the controller of the unit.
// PARAMETERS : player - pointer to new controller
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void Unit::SetOwner( Player *player ) {
  if ( player != u_player ) {
    if ( !IsDummy() ) {
      if ( u_player ) u_player->Units( -1 );
      player->Units( 1 );
    }
    u_pid = player->ID();
    u_player = player;
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::SetPosition
// DESCRIPTION: Change the position of the unit on the map. This is just
//              the low level function. It does not update the display
//              or do other fancy things. It's just about data shuffling.
// PARAMETERS : x - new x position
//              y - new y position
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void Unit::SetPosition( short x, short y ) {
  u_pos.x = x;
  u_pos.y = y;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Attack
// DESCRIPTION: Store an attack target. Again, this function does not
//              handle the display update, or the game mode change.
// PARAMETERS : enemy - unit to attack
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void Unit::Attack( const Unit *enemy ) {
  u_target = enemy->Position();
  SetFlags( U_ATTACKED|U_DONE );
  u_moves = 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::DefensiveStrength
// DESCRIPTION: Get a number measuring the defensive capabilities
//              of the unit.
// PARAMETERS : -
// RETURNS    : armour value
//
// HISTORY
////////////////////////////////////////////////////////////////////////

unsigned char Unit::DefensiveStrength( void ) const {
  return u_type->Armour();
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::OffensiveStrength
// DESCRIPTION: Return the combat value of the unit against the given
//              target.
// PARAMETERS : target   - unit to attack
// RETURNS    : offensive strength
//
// HISTORY
////////////////////////////////////////////////////////////////////////

unsigned char Unit::OffensiveStrength( const Unit *target ) const {
  unsigned char pow;

  if ( target->IsShip() || target->IsFloating() ) pow = u_type->Firepower(U_SHIP);
  else if ( target->IsAircraft() ) pow = u_type->Firepower(U_AIR);
  else pow = u_type->Firepower(U_GROUND);
  return pow;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Hit
// DESCRIPTION: Inflict damage to the unit and destroy it if necessary.
// PARAMETERS : damage - how much damage is done
// RETURNS    : true if the unit was destroyed, false otherwise
//
// HISTORY
//   20-05-2002 - check for dummy unit before decrementing unit count
////////////////////////////////////////////////////////////////////////

bool Unit::Hit( unsigned short damage ) {
  if ( IsAlive() ) {

    if ( u_group <= damage ) {     // unit destroyed
      u_group = 0;
      SetFlags( U_DESTROYED );
      u_pos.x = u_pos.y = -1;

      if ( !IsDummy() ) u_player->Units( -1 );
      return true;
    }

    u_group -= damage;
  }
  return false;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::CanHitType
// DESCRIPTION: Check whether the unit can shoot at an enemy unit. The
//              test is performed looking only at the enemy unit type
//              (ground, ship. aircraft, or submarine). If the function
//              returns true, this doesn't mean that the enemy is
//              actually in range for the unit's weapons systems.
// PARAMETERS : enemy - enemy unit
// RETURNS    : true if the unit type can be shot; false otherwise
//
// HISTORY
////////////////////////////////////////////////////////////////////////

bool Unit::CanHitType( const Unit *enemy ) const {
  return WeaponRange( enemy ) > 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::CanHit
// DESCRIPTION: Check whether the unit can shoot at an enemy unit. The
//              test is performed with regard to the enemy unit type
//              (ground, ship. aircraft, or submarine) and the distance
//              of the target.
// PARAMETERS : enemy - enemy unit
// RETURNS    : true if the unit can be shot; false if it's the wrong
//              type, too far away, or not hostile
//
// HISTORY
////////////////////////////////////////////////////////////////////////

bool Unit::CanHit( const Unit *enemy ) const {
  if ( (u_player == enemy->Owner()) ||
       (u_flags & (U_DONE|U_SHELTERED)) ||
       enemy->IsSheltered() ) return false;
  return CouldHit( enemy );
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::CouldHit
// DESCRIPTION: Check whether the unit could shoot at an enemy unit.
//              Current unit status is not taken into account, so this
//              function may return true, even though the unit already
//              attacked this turn or moved and can't attack any more.
// PARAMETERS : enemy - enemy unit
// RETURNS    : true if the unit type could be shot; false otherwise
//
// HISTORY
//   10-06-2002 - don't get confused with amphibian units
////////////////////////////////////////////////////////////////////////

bool Unit::CouldHit( const Unit *enemy ) const {
  unsigned short dist = Distance( u_pos, enemy->Position() );
  unsigned long type;

  // be careful with amphibian units
  if ( enemy->IsAircraft() ) type = U_AIR;
  else if ( enemy->IsShip() || enemy->IsFloating() ) type = U_SHIP;
  else type = U_GROUND;

  return u_type->IsInFOF( dist, type );
}

////////////////////////////////////////////////////////////////////
// NAME       : Unit::Repair
// DESCRIPTION: Repair the unit (completely). This will get the
//              group size of the unit back to the maximum, but will
//              also lower its experience by one point per two
//              "rookies".
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
//   17-06-2002 - removed parameter and return code
////////////////////////////////////////////////////////////////////////

void Unit::Repair( void ) {
  if ( u_group < MAX_GROUP_SIZE ) {
    u_xp = MAX( 0, u_xp - (MAX_GROUP_SIZE - u_group)/2 );
    u_group = MAX_GROUP_SIZE;
    u_moves = 0;
    SetFlags( U_DONE );	// can't move this turn
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::WeaponRange
// DESCRIPTION: Get the distance from which the unit may shoot at a
//              given target.
// PARAMETERS : u - target unit
// RETURNS    : maximum range of fire
//
// HISTORY
////////////////////////////////////////////////////////////////////////

unsigned char Unit::WeaponRange( const Unit *u ) const {
  unsigned char range;

  if ( u->IsAircraft() ) range = u_type->MaxFOF(U_AIR);
  else if ( u->IsShip() || u->IsFloating() )
    range = u_type->MaxFOF(U_SHIP);
  else range = u_type->MaxFOF(U_GROUND);

  return range;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::RefreshMoves
// DESCRIPTION: Get the unit back to full movement points.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void Unit::RefreshMoves( void ) {
  if ( IsMine() ) {
    if ( IsSheltered() ) u_moves = 2;
  } else u_moves = u_type->Speed();
}

