// CoMET - The Crimson Fields Map Editing Tool
// Copyright (C) 2002-2003 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

////////////////////////////////////////////////////////////////////////
// extwindow2.cpp
//
// History:
//  06-12-2002 - created
///////////////////////////////////////////////////////////////////////

#include <stdlib.h>

#include "extwindow2.h"
#include "extwindow.h"
#include "filewindow.h"
#include "fileio.h"
#include "uiaux.h"

const char *player2_labels[] = { "One", "Two", 0 };
const char *player3_labels[] = { "One", "Two", "Neutral", 0 };

// local static vars
static const char *dir_labels[] = { "North", "Northeast", "Southeast",
                                    "South", "Southwest", "Northwest", 0 };
extern const char *event_labels[];
extern const char *etrigger_labels[];

////////////////////////////////////////////////////////////////////////
// NAME       : ListExchangeWindow::ListExchangeWindow
// DESCRIPTION: This window displays to TLWList objects. The user can
//              move nodes from one list to the other and vice versa.
// PARAMETERS : l1     - first TLWList
//              l2     - second TLWList
//              label1 - label to be shown for l1
//              label2 - label to be shown for l2
//              view   - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

ListExchangeWindow::ListExchangeWindow( TLWList &l1, TLWList &l2,
            const char *label1, const char *label2, View *view )
          : Window(WIN_CENTER, view), list1(l1), list2(l2) {
  // calculate window size
  unsigned short maxlen = sfont->TextWidth(label1), nodes1 = 0, nodes2 = 0;
  TLWNode *n;

  for ( n = static_cast<TLWNode *>(l1.Head()); n;
        n = static_cast<TLWNode *>(n->Next()) ) {
    maxlen = MAX( maxlen, sfont->TextWidth(n->Name()) );
    ++nodes1;
  }
  for ( n = static_cast<TLWNode *>(l2.Head()); n;
        n = static_cast<TLWNode *>(n->Next()) ) {
    maxlen = MAX( maxlen, sfont->TextWidth(n->Name()) );
    ++nodes2;
  }

  Rect win( 0, 0, (maxlen * 2 + 10) + sfont->Width() + 30,
            (MAX(nodes1, nodes2) + 4) * (sfont->Height() + 4) + 30 );
  win.Clip( *view );
  SetSize( win );

  wd_l1 = new TextListWidget( 0, 5, sfont->Height() + 10,
          w/2 - sfont->Width() - 10, h - sfont->Height() * 2 - 20,
          &l1, -1, WIDGET_ALIGN_TOP, label1, this );

  wd_l2 = new TextListWidget( 0, w - 5 - wd_l1->Width(),
          wd_l1->TopEdge(), wd_l1->Width(), wd_l1->Height(),
          &l2, -1, WIDGET_ALIGN_TOP, label2, this );

  Widget *wd;
  wd = new ButtonWidget( B_ID_RIGHT,
       wd_l1->LeftEdge() + wd_l1->Width() + 5,
       wd_l1->TopEdge() + wd_l1->Height()/2 - sfont->Height() - 15,
       wd_l2->LeftEdge() - wd_l1->LeftEdge() - wd_l1->Width() - 10,
       sfont->Height() + 8, '>', 0, ">", this );
  wd->SetHook( this );

  wd = new ButtonWidget( B_ID_LEFT, wd->LeftEdge(),
       wd->TopEdge() + wd->Height() + 7, wd->Width(), wd->Height(),
       '<', 0, "<", this );
  wd->SetHook( this );

  ok = new ButtonWidget( B_ID_OK, 1, h - sfont->Height() - 9,
                         w - 2, sfont->Height() + 8, 'o',
                         WIDGET_DEFAULT, "OK", this );
  ok->SetHook( this );

  Draw();
  Show();
}

////////////////////////////////////////////////////////////////////////
// NAME       : ListExchangeWindow::ListExchangeWindow
// DESCRIPTION: React to user actions.
// PARAMETERS : widget - activated widget
//              win    - window containing the widget
// RETURNS    : GUI status
//
// HISTORY
////////////////////////////////////////////////////////////////////////

GUI_Status ListExchangeWindow::WidgetActivated( Widget *widget, Window *win ) {
  TLWNode *n;

  switch ( widget->ID() ) {
  case B_ID_RIGHT:
    n = static_cast<TLWNode *>(wd_l1->Selected());
    if ( n ) {
      n->Remove();
      list2.InsertNodeSorted( n );
      wd_l1->Update(); wd_l1->Draw(); wd_l1->Show();
      wd_l2->Update(); wd_l2->Draw(); wd_l2->Show();
    }
    break;
  case B_ID_LEFT:
    n = static_cast<TLWNode *>(wd_l2->Selected());
    if ( n ) {
      n->Remove();
      list1.InsertNodeSorted( n );
      wd_l1->Update(); wd_l1->Draw(); wd_l1->Show();
      wd_l2->Update(); wd_l2->Draw(); wd_l2->Show();
    }
    break;
  case B_ID_OK:
    view->CloseWindow( this );
    break;
  }

  return GUI_OK;
}

////////////////////////////////////////////////////////////////////////
// NAME       : NewMissionWindow::NewMissionWindow
// DESCRIPTION: This window pops up when the user wants to create a new
//              mission. He can adjust a few basic settings like map
//              size or tileset and create the mission.
// PARAMETERS : view - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

NewMissionWindow::NewMissionWindow( View *view ) : Window(WIN_CENTER, view) {

  // calculate window dimensions
  Font *f = SmallFont();
  unsigned short width = MIN( f->Width() * 38 + 58, view->Width() );
  unsigned short height = f->Height() * 3 + 40;
  unsigned short xoff = f->Width() * 11 + 14, yoff = 5;

  SetSize( width, height );

  m_width = new NumberWidget( 1, xoff, yoff, f->Width() * 4 + 10,
            f->Height() + 6, 'w', 40, 20, 200, WIDGET_ALIGN_LEFT,
            "Map Width :", this );

  width = MIN( f->Width() * 11 + 10, w - xoff - m_width->w - 24 - f->Width() * 12 );
  tileset = new StringWidget( 3, xoff + m_width->w + 10 + f->Width() * 9,
            yoff, width, m_width->h, 't', "default", 10,
            WIDGET_ALIGN_LEFT|WIDGET_STR_CONST, "Tile Set:", this );

  ButtonWidget *btn = new ButtonWidget( B_ID_TILESET,
             tileset->x + tileset->w, tileset->y,
             f->Width() * 3 + 4, tileset->h, 't', 0, "...", this );
  btn->SetHook( this );

  yoff += m_width->Height() + 5;

  m_height = new NumberWidget( 2, xoff, yoff, m_width->w, m_width->h,
            'h', 40, 20, 200, WIDGET_ALIGN_LEFT, "Map Height:", this );

  unitset = new StringWidget( 4, tileset->x, yoff, tileset->w, tileset->h,
            'u', "default", 10, WIDGET_ALIGN_LEFT|WIDGET_STR_CONST,
            "Unit Set:", this );

  btn = new ButtonWidget( B_ID_UNITSET,
             btn->x, unitset->y, btn->w, unitset->h, 'u', 0, "...", this );
  btn->SetHook( this );

  yoff = h - f->Height() - 15;
  ok = new ButtonWidget( B_ID_NMW_CREATE, 10, yoff, f->Width() * 10,
       f->Height() + 10, 'o', 0, "OK", this );

  new ButtonWidget( GUI_CLOSE, w - 10 - ok->w, yoff, ok->w, ok->h,
                    'c', 0, "Cancel", this );
  Draw();
  Show();
}

////////////////////////////////////////////////////////////////////////
// NAME       : NewMissionWindow::GetMapSize
// DESCRIPTION: Get the current slider settings.
// PARAMETERS : -
// RETURNS    : selected size as a Point
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Point NewMissionWindow::GetMapSize( void ) const {
  Point s( 0, 0 );

  s.x = m_width->Number();
  s.y = m_height->Number();

  return s;
}

////////////////////////////////////////////////////////////////////////
// NAME       : NewMissionWindow::WidgetActivated
// DESCRIPTION: React to user actions.
// PARAMETERS : button - activated button widget
//              win    - window containing the button
// RETURNS    : GUI status
//
// HISTORY
////////////////////////////////////////////////////////////////////////

GUI_Status NewMissionWindow::WidgetActivated( Widget *button, Window *win ) {
  FileWindow *fw;

  switch ( button->ID() ) {
  case B_ID_TILESET:
    fw = new FileWindow( get_data_dir().c_str(),
                         GetTileSet(), ".tiles", 0, view );
    fw->ok->SetID( B_ID_TILESET_OK );
    fw->ok->SetHook( this );
    break;
  case B_ID_TILESET_OK: {
    string file = static_cast<FileWindow *>(win)->GetFile();
    if ( file.length() > 0 ) {
      file = file_part( file );
      file.erase( file.length() - 6 );   // remove file name suffix
      tileset->SetString( file.c_str() );
    }
    break; }
  case B_ID_UNITSET:
    fw = new FileWindow( get_data_dir().c_str(),
                         GetUnitSet(), ".units", 0, view );
    fw->ok->SetID( B_ID_UNITSET_OK );
    fw->ok->SetHook( this );
    break;
  case B_ID_UNITSET_OK: {
    string file = static_cast<FileWindow *>(win)->GetFile();
    if ( file.length() > 0 ) {
      file = file_part( file );
      file.erase( file.length() - 6 );   // remove file name suffix
      tileset->SetString( file.c_str() );
    }
    break; }
  }

  return GUI_OK;
}

////////////////////////////////////////////////////////////////////////
// NAME       : EdBuildingWindow::EdBuildingWindow
// DESCRIPTION: This window contains the settings for a building.
// PARAMETERS : bld     - building to be modified
//              mission - currently active mission
//              view    - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

EdBuildingWindow::EdBuildingWindow( Building &bld, Mission &mission, View *view ) :
                  Window(WIN_CENTER, view), b(bld), mission(mission) { 
  // calculate dimensions
  SetSize( MIN(sfont->Width() * 50, view->Width()),
           MIN(sfont->Height() * 12, view->Height()) );

  unsigned short xoff, yoff = 5, wdh = sfont->Height() + 10;

  b_name = new StringWidget( B_ID_NAME, 5, yoff, w/2 - 10, wdh,
                             'n', b.Name(), 19, 0, NULL, this );

  xoff = w/2 + 10 + sfont->TextWidth("Player");
  b_player = new CycleWidget( B_ID_PLAYER, xoff, yoff, w - 5 - xoff, wdh,
                              'p', 0, "Player", b.Owner(), player3_labels, this );
  yoff += 5 + wdh;

  ButtonWidget *btn;
  btn = new CheckboxWidget( B_ID_WORKSHOP, 5, yoff, DEFAULT_CBW_SIZE,
                            DEFAULT_CBW_SIZE, b.IsWorkshop(), 'w',
                            WIDGET_STYLE_NOBORDER|WIDGET_STYLE_GFX|WIDGET_ALIGN_RIGHT,
                            "Workshop", this );
  btn->SetHook( this );

  btn = new CheckboxWidget( B_ID_FACTORY, 5, yoff + DEFAULT_CBW_SIZE + 5,
                            DEFAULT_CBW_SIZE, DEFAULT_CBW_SIZE, b.IsFactory(), 'f',
                            WIDGET_STYLE_NOBORDER|WIDGET_STYLE_GFX|WIDGET_ALIGN_RIGHT,
                            "Factory", this );
  btn->SetHook( this );

  btn = new ButtonWidget( B_ID_UNITS, w/2 + 5, yoff, (w - 20)/2, wdh,
                          'u', 0, "Units...", this ),
  btn->SetHook( this );
  yoff += 5 + wdh;

  b_factory_units = new ButtonWidget( B_ID_FACTORY_UNITS, btn->LeftEdge(), yoff,
                    btn->Width(), wdh, 'a', (b.IsFactory() ? 0 : WIDGET_DISABLED),
                    "Factory Settings...", this );
  b_factory_units->SetHook( this );
  yoff += 5 + wdh;

  xoff = 10 + sfont->TextWidth("Crystals");
  b_crystals = new NumberWidget( 0, xoff, yoff, sfont->Width() * 4 + 10, wdh,
                                 'c', b.Crystals(), 0, b.CrystalStore(),
                                 WIDGET_ALIGN_LEFT, "Crystals", this );

  xoff += b_crystals->Width() + 10 + sfont->TextWidth("of");
  b_maxcrystals = new NumberWidget( N_ID_MAXCRYSTALS, xoff, yoff,
                                  b_crystals->Width(), wdh, 0,
                                  b.CrystalStore(), 0, 1000,
                                  WIDGET_ALIGN_LEFT, "of", this );
  b_maxcrystals->SetHook( this );

  btn = new ButtonWidget( B_ID_OK, 1, h - 1 - wdh, w - 2, wdh, 'o',
                          WIDGET_DEFAULT, "OK", this ),
  btn->SetHook( this );

  Draw();
  Show();

  TLWListBuilder::BuildUnitTypesList( mission.GetUnitSet(), fact_list_na );
  if ( b.IsFactory() ) {
    TLWNode *n = static_cast<TLWNode *>(fact_list_na.Head()), *n2;

    while ( n ) {
      n2 = static_cast<TLWNode *>(n->Next());
      if ( b.CanProduce( n->ID() ) ) {
        n->Remove();
        fact_list_ok.InsertNodeSorted( n );
      }
      n = n2;
    }
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : EdBuildingWindow::WidgetActivated
// DESCRIPTION: React to user actions.
// PARAMETERS : widget - activated widget
//              win    - window containing the widget
// RETURNS    : GUI status
//
// HISTORY
////////////////////////////////////////////////////////////////////////

GUI_Status EdBuildingWindow::WidgetActivated( Widget *widget, Window *win ) {

  switch ( widget->ID() ) {
  case B_ID_WORKSHOP:
    if ( widget->Clicked() ) b.SetFlags( BLD_WORKSHOP );
    else b.UnsetFlags( BLD_WORKSHOP );
    break;
  case B_ID_FACTORY:
    if ( widget->Clicked() ) {
      b.SetFlags( BLD_FACTORY );
      b_factory_units->Enable();
    } else {
      b.UnsetFlags( BLD_FACTORY );
      b_factory_units->Disable();
    }
    b_factory_units->Draw();
    b_factory_units->Show();
    break;
  case B_ID_UNITS:
    new EdUnitsWindow( mission, b, view );
    break;
  case B_ID_FACTORY_UNITS:
    new ListExchangeWindow( fact_list_na, fact_list_ok,
                            "Not Available", "Available", view );
    break;
  case B_ID_OK:
    view->CloseWindow( win );
    b.SetOwner( b_player->GetValue() );
    b.SetName( b_name->String() );
    b.SetCrystals( b_crystals->Number() );
    b.SetMaxCrystals( b_maxcrystals->Number() );

    for ( Unit *u = static_cast<Unit *>(mission.GetUnits().Head()); u;
          u = static_cast<Unit *>(u->Next()) ) {
      if ( u->Position() == b.Position() ) u->SetOwner( b.Owner() );
    }

    // fix factory settings
    if ( b.IsFactory() ) {
      unsigned long unitmask = 0;

      for ( TLWNode *n = static_cast<TLWNode *>(fact_list_ok.Head());
            n; n = static_cast<TLWNode *>(n->Next()) ) {
        unitmask |= (1 << n->ID());
      }
      b.SetUnitProduction( unitmask );
    } else b.SetUnitProduction( 0 );
    break;
  case N_ID_MAXCRYSTALS:
    if ( b_crystals->Number() > b_maxcrystals->Number() )
      b_crystals->SetNumber( b_maxcrystals->Number() );
    break;
  }

  return GUI_OK;
}

////////////////////////////////////////////////////////////////////////
// NAME       : EdUnitWindow::EdUnitWindow
// DESCRIPTION: This window contains the settings for a unit.
// PARAMETERS : unit    - unit to be modified
//              mission - currently active mission
//              view    - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

EdUnitWindow::EdUnitWindow( Unit &unit, Mission &mission, View *view ) :
              Window(WIN_CENTER, view), u(unit), mission(mission), mv(0) { 
  // calculate dimensions
  SetSize( MIN(sfont->Width() * 50, view->Width()),
           MIN(sfont->Height() * 12, view->Height()) );

  unsigned short xoff, yoff = 5, wdh = sfont->Height() + 10;

  Widget *wd = new StringWidget( 0, 5, yoff, w/2 - 10, wdh,
                   0, u.Name(), 19,
                   WIDGET_STYLE_NOBORDER|WIDGET_STR_CONST,
                   NULL, this );
  xoff = w/2 + 10 + sfont->TextWidth("Player");
  u_player = new CycleWidget( B_ID_PLAYER, xoff, yoff, w - 5 - xoff, wdh,
                              'p', (u.IsSheltered() ? WIDGET_DISABLED : 0),
                              "Player", u.Owner(),
                              (u.IsSheltered() ? player3_labels : player2_labels),
                              this );
  u_player->SetHook( this );
  yoff += 5 + wdh;

  new NumberWidget( 0, wd->LeftEdge() + 10 + sfont->Width() * 2,
                    yoff, sfont->Width() * 6, wdh, 0, u.ID(), 0, 32000,
                    WIDGET_ALIGN_LEFT|WIDGET_STYLE_NOBORDER|WIDGET_STR_CONST,
                    "ID", this );

  xoff = w/2 + 10 + sfont->TextWidth("Direction");
  wd = new CycleWidget( B_ID_DIRECTION, xoff, yoff, w - 5 - xoff, wdh,
                        'd', 0, "Direction", u.GetDirection(), dir_labels, this );
  wd->SetHook( this );
  yoff += 5 + wdh;

  if ( u.IsTransport() ) {
    xoff = 10 + sfont->TextWidth("Crystals");
    // maximum number of crystals should be
    // (mission.StorageLeft(u) + (u.Crystals() + 9)/10) * 10,
    // but we can't update that maxium when the list of contained units
    // changes, so we allow the maximum here and update manually when the
    // widget is activated if necessary
    wd = new NumberWidget( N_ID_CRYSTALS, xoff, yoff, sfont->Width() * 6, wdh,
                           0, u.Crystals(), 0, u.Type()->Slots() * 10,
                           WIDGET_ALIGN_LEFT, "Crystals", this );
    wd->SetHook( this );

    if ( !u.IsSheltered() ) {
      wd = new ButtonWidget( B_ID_UNITS, w/2 + 5, yoff, (w - 20)/2, wdh,
                             'u', 0, "Units...", this ),
      wd->SetHook( this );
    }
  }

  wd = new ButtonWidget( B_ID_OK, 1, h - 1 - wdh, w - 2, wdh,
                         'o', WIDGET_DEFAULT, "OK", this ),
  wd->SetHook( this );

  Draw();
  Show();
}

////////////////////////////////////////////////////////////////////////
// NAME       : EdUnitWindow::WidgetActivated
// DESCRIPTION: React to user actions.
// PARAMETERS : widget - activated widget
//              win    - window containing the widget
// RETURNS    : GUI status
//
// HISTORY
////////////////////////////////////////////////////////////////////////

GUI_Status EdUnitWindow::WidgetActivated( Widget *widget, Window *win ) {
  GUI_Status rc = GUI_OK;

  switch ( widget->ID() ) {
  case B_ID_DIRECTION:
    u.SetDirection( static_cast<CycleWidget *>(widget)->GetValue() );
    break;
  case B_ID_UNITS:
    new EdUnitsWindow( mission, u, view );
    break;
  case N_ID_CRYSTALS: {
    NumberWidget *cwidget = static_cast<NumberWidget *>(widget);
    unsigned short crystals = MIN( cwidget->Number(),
                  (mission.StorageLeft(u) + (u.Crystals() + 9)/10) * 10 );
    if ( cwidget->Number() != crystals ) cwidget->SetNumber( crystals );
    u.SetCrystals( crystals );
    break; }
  case B_ID_OK:
    view->CloseWindow( win );
    u.SetOwner( u_player->GetValue() );

    if ( u.IsTransport() && !u.IsSheltered() ) {
      for ( Unit *walk = static_cast<Unit *>(mission.GetUnits().Head());
            walk; walk = static_cast<Unit *>(walk->Next()) ) {
        if ( walk->Position() == u.Position() ) walk->SetOwner( u.Owner() );
      }
    }

    if ( mv ) view->Refresh( mv->UpdateHex(u.Position()) );
    break;
  }

  return rc;
}

////////////////////////////////////////////////////////////////////////
// NAME       : EdEventsWindow::EdEventsWindow
// DESCRIPTION: This window presents the list of events for selection.
// PARAMETERS : mission - current mission
//              view    - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

EdEventsWindow::EdEventsWindow( Mission &mission, View *view ) :
              Window(WIN_CENTER, view), mission(mission) { 
  // calculate dimensions
  SetSize( MIN(sfont->Width() * 20 + 20, view->Width()),
           MIN( MAX(mission.GetEvents().CountNodes() + 2, 6) *
                   (sfont->Height() + 2) + 40,
                view->Height()) );

  TLWListBuilder::BuildEventList( mission.GetEvents(), nodes );

  e_list = new TextListWidget( 0, 5, 5, w - 10, h - sfont->Height() * 2 - 30,
               &nodes, -1, WIDGET_HSCROLLKEY|WIDGET_VSCROLLKEY, NULL, this );

  ButtonWidget *btn = new ButtonWidget( B_ID_NEW,
               1, h - sfont->Height() * 2 - 17, (w - 2)/2, sfont->Height() + 8,
               'n', 0, "New", this );
  btn->SetHook( this );

  btn = new ButtonWidget( B_ID_EDIT,
              btn->LeftEdge() + btn->Width(), btn->TopEdge(),
              btn->Width(), btn->Height(), 'e', 0, "Edit", this );
  btn->SetHook( this );

  btn = new ButtonWidget( B_ID_DELETE,
              btn->LeftEdge() - btn->Width(), btn->TopEdge() + btn->Height(),
              btn->Width(), btn->Height(), 'd', 0, "Delete", this );
  btn->SetHook( this );

  btn = new ButtonWidget( B_ID_OK,
              btn->LeftEdge() + btn->Width(), btn->TopEdge(),
              btn->Width(), btn->Height(), 'o', 0, "OK", this );
  btn->SetHook( this );

  Draw();
  Show();
}

////////////////////////////////////////////////////////////////////////
// NAME       : EdEventsWindow::WidgetActivated
// DESCRIPTION: React to user actions.
// PARAMETERS : button - activated button widget
//              win    - window containing the button
// RETURNS    : GUI status
//
// HISTORY
////////////////////////////////////////////////////////////////////////

GUI_Status EdEventsWindow::WidgetActivated( Widget *button, Window *win ) {

  switch ( button->ID() ) {
  case B_ID_EDIT:
    new NoteWindow("Sorry", "Not yet implemented", 0, view); // FIXME
    break;
  case B_ID_NEW:
    view->CloseWindow( win );
    new SelectEventWindow( mission, view );
    break;
  case B_ID_DELETE: {
    TLWNode *n = static_cast<TLWNode *>( e_list->Selected() );
    if ( n ) {
      Event *e = (Event *)n->UserData();

      n->Remove();
      delete n;
      e_list->Update();
      e_list->Draw();
      e_list->Show();

      mission.DeleteEvent( e );
    }
    break; }
  case B_ID_OK:
    view->CloseWindow( win );
    break;
  }

  return GUI_OK;
}


////////////////////////////////////////////////////////////////////////
// NAME       : SelectEventWindow::SelectEventWindow
// DESCRIPTION: This window presents the list of event types.
// PARAMETERS : mission - current mission
//              view    - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

SelectEventWindow::SelectEventWindow( Mission &mission, View *view ) :
                   Window(WIN_CENTER, view), mission(mission) { 
  // calculate dimensions
  SetSize( MIN(sfont->Width() * 18 + 20, view->Width()),
           MIN( 10 * (sfont->Height() + 2) + 42, view->Height()) );

  BuildTLWList( nodes );

  e_list = new TextListWidget( 0, 5, 5, w - 10, h - sfont->Height() * 3 - 36,
               &nodes, -1, WIDGET_HSCROLLKEY|WIDGET_VSCROLLKEY, NULL, this );

  e_trigger = new CycleWidget( B_ID_TRIGGER, e_list->LeftEdge(),
                  e_list->TopEdge() + e_list->Height() + sfont->Height() + 10,
                  e_list->Width(), sfont->Height() + 8, 't', WIDGET_ALIGN_TOP,
                  "Triggered by", 0, etrigger_labels, this );

  ButtonWidget *btn = new ButtonWidget( B_ID_OK,
               1, h - sfont->Height() - 9, (w - 2)/2, sfont->Height() + 8,
               'o', WIDGET_DEFAULT, "OK", this );
  btn->SetHook( this );

  btn = new ButtonWidget( B_ID_CANCEL,
              btn->LeftEdge() + btn->Width(), btn->TopEdge(),
              btn->Width(), btn->Height(), 'c', 0, "Cancel", this );
  btn->SetHook( this );

  Draw();
  Show();
}

////////////////////////////////////////////////////////////////////////
// NAME       : SelectEventWindow::WidgetActivated
// DESCRIPTION: React to user actions.
// PARAMETERS : button - activated button widget
//              win    - window containing the button
// RETURNS    : GUI status
//
// HISTORY
////////////////////////////////////////////////////////////////////////

GUI_Status SelectEventWindow::WidgetActivated( Widget *button, Window *win ) {

  switch ( button->ID() ) {
  case B_ID_OK:

/*** FIXME
    if ( e_list->Selected() ) {
      view->CloseWindow( win );

      Event *e = mission.CreateEvent(
                     static_cast<TLWNode *>(e_list->Selected())->ID(),
                     e_trigger->GetValue() );

      switch ( e->Type() ) {
      case EVENT_CREATE_UNIT:      new EventCreateUnitWindow( e ); break;
      case EVENT_MANIPULATE_EVENT: new EventManipulateEventWindow( e ); break;
      case EVENT_MESSAGE:          new EventMessageWindow( e ); break;
      case EVENT_MINING:           new EventMiningWindow( e ); break;
      case EVENT_NEXT_MAP:         new EventNextMapWindow( e ); break;
      case EVENT_RESEARCH:         new EventResearchWindow( e ); break;
      case EVENT_SCORE:            new EventScoreWindow( e ); break;
      }
    }
***/

    new NoteWindow("Sorry", "Not yet implemented", 0, view);
    break;
  case B_ID_CANCEL:
    view->CloseWindow( win );
    break;
  }

  return GUI_OK;
}

////////////////////////////////////////////////////////////////////////
// NAME       : SelectEventWindow::BuildTLWList
// DESCRIPTION: Create a list for a TextListWidget containing all
//              available event types.
// PARAMETERS : list - list of TLWNodes to be filled
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void SelectEventWindow::BuildTLWList( TLWList &events ) const {
  for ( int i = 0; event_labels[i]; ++i )
    events.InsertNodeSorted( new TLWNode(event_labels[i], 0, i+1) );
}


////////////////////////////////////////////////////////////////////////
// NAME       : EdUnitsWindow::EdUnitsWindow
// DESCRIPTION: This window presents the list of units in a unit
//              container (building or transporter) for selection.
// PARAMETERS : mission   - current mission
//              container - selected unit container
//              view      - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

EdUnitsWindow::EdUnitsWindow( Mission &mission, MapObject &container, View *view ) :
              Window(WIN_CENTER, view), mission(mission), container(container) { 
  TLWListBuilder::BuildUnitList( mission.GetUnits(), nodes, container.Position() );

  // calculate dimensions
  SetSize( MIN(sfont->Width() * 30 + 20, view->Width()),
           MIN( MAX(nodes.CountNodes() + 2, 6) *
                   (sfont->Height() + 2) + 40,
                view->Height()) );

  e_list = new TextListWidget( 0, 5, 5, w - 10, h - sfont->Height() * 2 - 30,
               &nodes, -1, WIDGET_HSCROLLKEY|WIDGET_VSCROLLKEY, NULL, this );

  ButtonWidget *btn = new ButtonWidget( B_ID_NEW,
               1, h - sfont->Height() * 2 - 17, (w - 2)/2, sfont->Height() + 8,
               'n', 0, "New", this );
  btn->SetHook( this );

  btn = new ButtonWidget( B_ID_EDIT,
              btn->LeftEdge() + btn->Width(), btn->TopEdge(),
              btn->Width(), btn->Height(), 'e', 0, "Edit", this );
  btn->SetHook( this );

  btn = new ButtonWidget( B_ID_DELETE,
              btn->LeftEdge() - btn->Width(), btn->TopEdge() + btn->Height(),
              btn->Width(), btn->Height(), 'd', 0, "Delete", this );
  btn->SetHook( this );

  btn = new ButtonWidget( B_ID_OK,
              btn->LeftEdge() + btn->Width(), btn->TopEdge(),
              btn->Width(), btn->Height(), 'o', 0, "OK", this );
  btn->SetHook( this );

  Draw();
  Show();
}

////////////////////////////////////////////////////////////////////////
// NAME       : EdUnitsWindow::WidgetActivated
// DESCRIPTION: React to user actions.
// PARAMETERS : button - activated button widget
//              win    - window containing the button
// RETURNS    : GUI status
//
// HISTORY
////////////////////////////////////////////////////////////////////////

GUI_Status EdUnitsWindow::WidgetActivated( Widget *button, Window *win ) {
  TLWNode *n;

  switch ( button->ID() ) {
  case B_ID_EDIT:
    n = static_cast<TLWNode *>( e_list->Selected() );
    if ( n ) new EdUnitWindow( *((Unit *)n->UserData()), mission, view );
    break;
  case B_ID_NEW:
    utypes.Clear();
    BuildTLWList( utypes );
    new SelectUnitWindow( mission, utypes, this, view );
    break;
  case B_ID_DELETE:
    n = static_cast<TLWNode *>( e_list->Selected() );
    if ( n ) {
      Unit *u = (Unit *)n->UserData();

      n->Remove();
      delete n;
      e_list->Update();
      e_list->Draw();
      e_list->Show();

      mission.DeleteUnit( u );
    }
    break;
  case B_ID_OK:
    view->CloseWindow( win );
    break;

  case B_ID_SUW_OK: {
    const UnitType *ut = static_cast<SelectUnitWindow *>(win)->Selected();
    view->CloseWindow( win );

    if ( ut ) {
      Unit *u = mission.CreateUnit( ut, container.Owner(), container.Position() );
      if ( u ) {
        char buf[28];
        sprintf( buf, "%s (%d)", u->Name(), u->ID() );
        nodes.InsertNodeSorted( new TLWNode( buf, u ) );
        e_list->Update();
        e_list->Draw();
        e_list->Show();
      }
    }
    break; }
  }

  return GUI_OK;
}

////////////////////////////////////////////////////////////////////////
// NAME       : EdUnitsWindow::BuildTLWList
// DESCRIPTION: Construct a list with all legal unit types for the
//              current container.
// PARAMETERS : list - list to add nodes to
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void EdUnitsWindow::BuildTLWList( TLWList &list ) const {
  const UnitSet &us = mission.GetUnitSet();
  TLWListBuilder::BuildUnitTypesList( us, list );

  // now remove all unit types which are not allowed by the container
  unsigned short maxw = container.MaxWeight();
  if ( container.IsUnit() )
    maxw = MIN( maxw, mission.StorageLeft(static_cast<Unit &>(container)) );

  TLWNode *n = static_cast<TLWNode *>(list.Head()), *n2;
  while ( n ) {

    const UnitType *ut = us.GetUnitInfo( n->ID() );
    if ( (ut->Weight() < container.MinWeight()) ||
         (ut->Weight() > maxw) ) {
      n2 = static_cast<TLWNode *>(n->Next());
      n->Remove();
      delete n;
      n = n2;
    } else n = static_cast<TLWNode *>(n->Next());
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : SelectUnitWindow::SelectUnitWindow
// DESCRIPTION: This window presents the list of unit types.
// PARAMETERS : mission - current mission
//              hook    - widget hook to activate when OK is pressed
//              view    - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

SelectUnitWindow::SelectUnitWindow( Mission &mission,
                  WidgetHook *hook, View *view ) :
                   Window(WIN_CENTER, view), mission(mission) {
  TLWListBuilder::BuildUnitTypesList( mission.GetUnitSet(), nodes );

  Init( nodes, hook );
}


////////////////////////////////////////////////////////////////////////
// NAME       : SelectUnitWindow::SelectUnitWindow
// DESCRIPTION: This window presents the list of unit types.
// PARAMETERS : mission - current mission
//              list    - use this list instead of constructing our own
//              hook    - widget hook to activate when OK is pressed
//              view    - pointer to the window's view
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

SelectUnitWindow::SelectUnitWindow( Mission &mission, TLWList &list,
                  WidgetHook *hook, View *view ) :
                   Window(WIN_CENTER, view), mission(mission) {
  Init( list, hook  );
}

////////////////////////////////////////////////////////////////////////
// NAME       : SelectUnitWindow::Init
// DESCRIPTION: Set up the window properties and create the widgets.
// PARAMETERS : list - list to present
//              hook - widget hook to activate when OK is pressed
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void SelectUnitWindow::Init( TLWList &list, WidgetHook *hook ) {
  // calculate dimensions
  SetSize( MIN(sfont->Width() * 24 + 20, view->Width()),
           MIN( MAX(list.CountNodes() + 1, 6) *
                   (sfont->Height() + 2) + 25,
                view->Height()) );

  e_list = new TextListWidget( 0, 5, 5, w - 10, h - sfont->Height() - 15,
               &list, -1, WIDGET_HSCROLLKEY|WIDGET_VSCROLLKEY, NULL, this );

  ButtonWidget *btn = new ButtonWidget( B_ID_SUW_OK,
               1, h - sfont->Height() - 9, (w - 2)/2, sfont->Height() + 8,
               'o', WIDGET_DEFAULT, "OK", this );
  btn->SetHook( hook );

  btn = new ButtonWidget( B_ID_CANCEL,
              btn->LeftEdge() + btn->Width(), btn->TopEdge(),
              btn->Width(), btn->Height(), 'c', 0, "Cancel", this );
  btn->SetHook( this );

  Draw();
  Show();
}

////////////////////////////////////////////////////////////////////////
// NAME       : SelectUnitWindow::WidgetActivated
// DESCRIPTION: React to user actions.
// PARAMETERS : button - activated button widget
//              win    - window containing the button
// RETURNS    : GUI status
//
// HISTORY
////////////////////////////////////////////////////////////////////////

GUI_Status SelectUnitWindow::WidgetActivated( Widget *button, Window *win ) {

  if ( button->ID() == B_ID_CANCEL ) view->CloseWindow( win );

  return GUI_OK;
}

////////////////////////////////////////////////////////////////////////
// NAME       : SelectUnitWindow::Selected
// DESCRIPTION: Get the selected unit type.
// PARAMETERS : -
// RETURNS    : selected unit type or NULL if none selected
//
// HISTORY
////////////////////////////////////////////////////////////////////////

const UnitType *SelectUnitWindow::Selected( void ) const {
  TLWNode *n = static_cast<TLWNode *>(e_list->Selected());

  if ( n ) return mission.GetUnitSet().GetUnitInfo( n->ID() );
  return NULL;
}

