// CoMET - The Crimson Fields Map Editing Tool
// Copyright (C) 2002-2003 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

/////////////////////////////////////////////////////////////////////////
// unit.cpp
//
// History:
//  09-11-2002 - created
////////////////////////////////////////////////////////////////////////

#include "SDL_endian.h"

#include "unit.h"
#include "hexsup.h"

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Unit
// DESCRIPTION: Create a new unit.
// PARAMETERS : type - pointer to a unit type definition
//              pid  - owner identifier
//              id   - unique unit identifier
//              pos  - position on map
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Unit::Unit( const UnitType *type, unsigned char pid, unsigned short id, const Point &pos ) :
      MapObject( MO_UNIT, type->Name() ) {
  u_type = type;

  u_pos = pos;
  u_flags = type->Flags();
  u_id = id;

  u_moves = type->Speed();
  u_facing = (pid == PLAYER_TWO) ? SOUTH : NORTH;
  u_group = MAX_GROUP_SIZE;
  u_xp = 0;

  u_pid = pid;
  u_crystals = 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Unit
// DESCRIPTION: Load a unit from a data file.
// PARAMETERS : file   - descriptor of an open data file
//              type   - pointer to a unit type definition
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Unit::Unit( SDL_RWops *file, const UnitType *type ) :
      MapObject( MO_UNIT, type->Name() ) {
  SDL_RWread( file, &u_pid, 1, 1 );
  u_pos.x = SDL_ReadLE16( file );
  u_pos.y = SDL_ReadLE16( file );
  u_flags = SDL_ReadLE32( file );
  u_id = SDL_ReadLE16( file );

  SDL_RWread( file, &u_moves, 1, 4 );

  SDL_ReadLE32( file );   // ignore target information

  if ( IsTransport() ) u_crystals = SDL_ReadLE16( file );
  else u_crystals = 0;

  u_type = type;
}


////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Save
// DESCRIPTION: Save the unit to a data file.
// PARAMETERS : file - descriptor of the save file
// RETURNS    : 0 on succes, non-zero on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int Unit::Save( SDL_RWops *file ) const {
  unsigned char typeno = u_type->ID();
  SDL_RWwrite( file, &typeno, 1, 1 );
  SDL_RWwrite( file, &u_pid, 1, 1 );

  SDL_WriteLE16( file, u_pos.x );
  SDL_WriteLE16( file, u_pos.y );
  SDL_WriteLE32( file, u_flags );
  SDL_WriteLE16( file, u_id );

  SDL_RWwrite( file, &u_moves, 1, 4 );

  SDL_WriteLE16( file, (unsigned short)-1 );  // target information
  SDL_WriteLE16( file, (unsigned short)-1 );

  if ( IsTransport() ) SDL_WriteLE16( file, u_crystals );

  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Unit::Export
// DESCRIPTION: Save the unit to a plain text file.
// PARAMETERS : file - descriptor of the save file
// RETURNS    : 0 on succes, non-zero on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int Unit::Export( ofstream &file ) const {
  file << "[unit]\n";
  file << "type = " << Name() << '\n';
  file << "player = " << (u_pid+1)%3 << '\n';
  file << "id = " << u_id << '\n';
  file << "xpos = " << u_pos.x << '\n';
  file << "ypos = " << u_pos.y << '\n';
  file << "face = " << (short)u_facing << '\n';
  if ( IsTransport() ) file << "crystals = " << u_crystals << '\n';
  file << '\n';
  return 0;
}

