#include <inklevel.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <linux/config.h>

#include <ieee1284.h>

#define BUFLEN 1024
#define NR_TAGS 15

#define IOCNR_GET_DEVICE_ID 1
#define LPIOC_GET_DEVICE_ID _IOC(_IOC_READ, 'P', IOCNR_GET_DEVICE_ID, BUFLEN)

static int get_device_id(int, int, unsigned char *);
static int parse_device_id(const unsigned char *, struct ink_level *);
static int parse_device_id_new_hp(unsigned char [NR_TAGS][BUFLEN], int, 
				  struct ink_level *);
static int parse_device_id_old_hp(unsigned char [NR_TAGS][BUFLEN], int, 
				  struct ink_level *);
static int my_axtoi(unsigned char *);
static int my_atoi(unsigned char *);

int get_ink_level(int port, int portnumber, struct ink_level *level) {
  unsigned char device_id[BUFLEN];

  *level = (struct ink_level) { NO_INK_LEVEL_FOUND, 0, 0, 0, 0, 0, 0, 0, 0,};
  
  if (get_device_id(port, portnumber, device_id) == OK) {
    if (parse_device_id(device_id, level) == OK) {
      return OK;
    }
  }

  return ERROR;
}

/* This function retrieves the device id of the specified port */

static int get_device_id(int port, int portnumber, unsigned char *device_id) {
  struct parport_list parports;
  unsigned char tmp[BUFLEN];
  unsigned char device_file[256];
  int size;
  int fd;

  if (port == PARPORT ) {
    if (ieee1284_find_ports(&parports, 0) == E1284_OK) {
      if (portnumber < parports.portc) {
	size = ieee1284_get_deviceid(parports.portv[portnumber], -1, 
				      F1284_FRESH, tmp, BUFLEN);
	strncpy(device_id, tmp + 2, size - 2);
	return OK;
      }
    }
    return ERROR;
  } else if (port == USB) {
#ifdef CONFIG_DEVFS_FS    
    sprintf(device_file, "/dev/usb/lp%d", portnumber);
#else
    sprintf(device_file, "/dev/usblp%d", portnumber);
#endif
    fd = open(device_file, O_RDONLY);
    if (fd == ERROR) {
      return ERROR;
    }
    if (ioctl(fd, LPIOC_GET_DEVICE_ID, tmp) < 0) {
      close(fd);
      return ERROR;
    }
    close(fd);
    size = (tmp[0] << 8) | (tmp[1]);
    size = (size < BUFLEN - 1) ? size : BUFLEN - 1;
    tmp[size] = '\0';
    strncpy(device_id, tmp + 2, size - 2);
    return OK;
  } else {
    return ERROR;
  }
}

/* This function parses the device id and calls the appropiate function */

static int parse_device_id(const unsigned char *device_id, 
			   struct ink_level *level) {
  int i = 0;
  int j = 0;
  const unsigned char *c;
  unsigned char tags[NR_TAGS][BUFLEN];  

  /* init tags */

  for (i = 0; i < NR_TAGS; i++) {
    for (j = 0; j < BUFLEN; j++) {
      tags[i][j] = '\0';
    }
  }

  /* Tokenize the device id */

  i = 0;
  j = 0;
  c = device_id;
  
  while ((*c != '\0') && (i < NR_TAGS)) {
    j = 0;
    while ((*c != ';') && (j < BUFLEN)) {
      tags[i][j] = *c;
      c++;
      j++;
    }
    c++; /* Skip the ';' */
    i++;
  }

#ifdef DEBUG
  /* print all tags */

  for (i = 0; i < NR_TAGS; i++) {
    printf("%d: %s\n", i, tags[i]);
  }
#endif

  /* Check if we deal with a printer */

  /* Find the "CLS:" tag */

  for (i = 0; i < NR_TAGS; i++) {
    c = tags[i];
    if ((c[0] == 'C') && (c[1] == 'L') && (c[2] == 'S') && (c[3] == ':')) {
      break;
    }
  }

  if (i < NR_TAGS) {

    /* Check if it is "PRINTER" */

    if ((tags[i][4] != 'P') || (tags[i][5] != 'R') || (tags[i][6] != 'I') || 
	(tags[i][7] != 'N') || (tags[i][8] != 'T') || (tags[i][9] != 'E') || 
	(tags[i][10] != 'R')) {
      
#ifdef DEBUG
      printf("No printer found\n");
#endif
      
      return ERROR;
    }
  } else { 
    /* Find the "CLASS:" tag */
    
    for (i = 0; i < NR_TAGS; i++) {
      c = tags[i];
      if ((c[0] == 'C') && (c[1] == 'L') && (c[2] == 'A') && (c[3] == 'S') && 
	  (c[4] == 'S') && (c[5] == ':')) {
	break;
      }
    }

    if (i < NR_TAGS) {
    
      /* Check if it is "PRINTER" */
      
      if ((tags[i][6] != 'P') || (tags[i][7] != 'R') || (tags[i][8] != 'I') || 
	  (tags[i][9] != 'N') || (tags[i][10] != 'T') || 
	  (tags[i][11] != 'E') || (tags[i][12] != 'R')) {
      
#ifdef DEBUG
	printf("No printer found\n");
#endif
	
	return ERROR;
      }
    } else {

#ifdef DEBUG 
      printf("No device class found\n");
#endif

      return ERROR;
    }
  }

  /* Check for a new HP printer */

  /* Find the "S:" tag */

  for (i = 0; i < NR_TAGS; i++) {
    c = tags[i];
    if ((c[0] == 'S') && (c[1] == ':')) {
      break;
    }
  }

#ifdef DEBUG
  if (i < NR_TAGS) {
    printf("The \"S:\" tag has number %d\n", i);
  } else {
    printf("No \"S:\" tag found\n");
  }
#endif

  if (i<NR_TAGS) {
    return parse_device_id_new_hp(tags, i, level);
  }
  
  /* Check for an old HP printer */

  /* Find the "VSTATUS:" tag */

  for (i = 0; i < NR_TAGS; i++) {
    c = tags[i];
    if ((c[0] == 'V') && (c[1] == 'S') && (c[2] == 'T') && (c[3] == 'A') &&
	(c[4] == 'T') && (c[5] == 'U') && (c[6] == 'S') && (c[7] ==':')) {
      break;
    }
  }

#ifdef DEBUG
  if (i < NR_TAGS) {
    printf("The \"VSTATUS:\" tag has number %d\n", i);
  } else {
    printf("No \"VSTATUS:\" tag found\n");
  }
#endif

  if (i<NR_TAGS) {
    return parse_device_id_old_hp(tags, i, level);
  }

  /* Insert code to check for other printers here */
  
  return ERROR; /* No matching printer was found */
}

/* This function parses the device id of a new HP printer
 * for example HP Deskjet 5550 
 */

static int parse_device_id_new_hp(unsigned char tags[NR_TAGS][BUFLEN], int n, 
				  struct ink_level *level) {
  unsigned char *s = tags[n];
  int length = 0;
  unsigned char b[3]; /* level of black ink as hexadecimal string */
  unsigned char c[3]; /* level of color ink as hexadecimal string */


  /* Determine the length of the string */

  while (*s != '\0') {
    length++;
    s++;
  }

  s = tags[n];
  
  b[0] = s[length - 10];
  b[1] = s[length -9];
  b[2] = '\0';

  c[0] = s[length - 2];
  c[1] = s[length - 1];
  c[2] = '\0';

  level->black = my_axtoi(b);
  level->color = my_axtoi(c);
  level->type = TWO_COLORS_FOUND;

#ifdef DEBUG
  printf("Black: %d%%, Color: %d%%\n", level->black, level->color);
#endif

  return OK;
}

/* This function parses the device id of an old HP printer
 * for example HP Photosmart 1000 
 */

static int parse_device_id_old_hp(unsigned char tags[NR_TAGS][BUFLEN], int n, 
				  struct ink_level *level) {
  unsigned char *s = tags[n];
  int length = 0;
  unsigned char b[4]; /* level of black ink as decimal string */
  unsigned char c[4]; /* level of color ink as decimal string */

  /* Determine the length of the string */

  while (*s != '\0') {
    length++;
    s++;
  }

  s = tags[n];
  
  /* Check if "KP" and "CP" tags are there */

  if ((s[length - 11] == 'K') && (s[length - 10] == 'P') &&
      (s[length - 5]  == 'C') && (s[length - 4] == 'P')) {
    b[0] = s[length - 9];
    b[1] = s[length - 8];
    b[2] = s[length - 7];
    b[3] = '\0';
    
    c[0] = s[length - 3];
    c[1] = s[length - 2];
    c[2] = s[length - 1];
    c[3] = '\0';

    level->black = my_atoi(b);
    level->color = my_atoi(c);
    level->type = TWO_COLORS_FOUND;

#ifdef DEBUG
    printf("Black; %d%%, Color %d%%\n", level->black, level->color);
#endif

    return OK;
  } else {

#ifdef DEBUG
    printf("No ink level found\n");
#endif
    return ERROR;
  }
}

/* This function converts a string containing a two digit hexadecimal number
 * to an int 
 */

static int my_axtoi(unsigned char* t) {
  int r = 0;

  switch (t[0]) {
  case '0':
    r = 0;
    break;
  case '1':
    r = 16;
    break;
  case '2':
    r = 32;
    break;
  case '3':
    r = 48;
    break;
  case '4':
    r = 64;
    break;
  case '5':
    r = 80;
    break;
  case '6':
    r = 96;
    break;
  case '7':
    r = 112;
    break;
  case '8':
    r = 128;
    break;
  case '9':
    r = 144;
    break;
  case 'a':
  case 'A':
    r = 160;
    break;
  case 'b':
  case 'B':
    r = 176;
    break;
  case 'c':
  case 'C':
    r = 192;
    break;
  case 'd':
  case 'D':
    r = 208;
    break;
  case 'e':
  case 'E':
    r = 224;
    break;
  case 'f':
  case 'F':
    r = 240;
    break;
  }

  switch (t[1]) {
  case '0':
    r += 0;
    break;
  case '1':
    r += 1;
    break;
  case '2':
    r += 2;
    break;
  case '3':
    r += 3;
    break;
  case '4':
    r += 4;
    break;
  case '5':
    r += 5;
    break;
  case '6':
    r += 6;
    break;
  case '7':
    r += 7;
    break;
  case '8':
    r += 8;
    break;
  case '9':
    r += 9;
    break;
  case 'a':
  case 'A':
    r += 10;
    break;
  case 'b':
  case 'B':
    r += 11;
    break;
  case 'c':
  case 'C':
    r += 12;
    break;
  case 'd':
  case 'D':
    r += 13;
    break;
  case 'e':
  case 'E':
    r += 14;
    break;
  case 'f':
  case 'F':
    r += 15;
    break;
  }

  return r;
}

/* This function converts a string containig a three digit decimal number
 * to an int 
 */

static int my_atoi(unsigned char* t) {
  int r = 0;

  switch (t[0]) {
  case '0':
    r = 0;
    break;
  case '1':
    r = 100;
    break;
  case '2':
    r = 200;
    break;
  case '3':
    r = 300;
    break;
  case '4':
    r = 400;
    break;
  case '5':
    r = 500;
    break;
  case '6':
    r = 600;
    break;
  case '7':
    r = 700;
    break;
  case '8':
    r = 800;
    break;
  case '9':
    r = 900;
    break;
  }
  
  switch (t[1]) {
  case '0':
    r += 0;
    break;
  case '1':
    r += 10;
    break;
  case '2':
    r += 20;
    break;
  case '3':
    r += 30;
    break;
  case '4':
    r += 40;
    break;
  case '5':
    r += 50;
    break;
  case '6':
    r += 60;
    break;
  case '7':
    r += 70;
    break;
  case '8':
    r += 80;
    break;
  case '9':
    r += 90;
    break;
  }

  switch (t[2]) {
  case '0':
    r += 0;
    break;
  case '1':
    r += 1;
    break;
  case '2':
    r += 2;
    break;
  case '3':
    r += 3;
    break;
  case '4':
    r += 4;
    break;
  case '5':
    r += 5;
    break;
  case '6':
    r += 6;
    break;
  case '7':
    r += 7;
    break;
  case '8':
    r += 8;
    break;
  case '9':
    r += 9;
    break;
  }
  
  return r;
}
